//
//  Mark Nelson
//  March 8, 1996
//  modified by Stefano Lonardi
//  September, 1998
//
//  This program performs an order-0 adaptive arithmetic encoding
//  function on an input file/stream, and sends the result to an
//  output file or stream.
//
//  This program contains the source code from the 1987 CACM
//  article by Witten, Neal, and Cleary.  I have taken the
//  source modules and combined them into this single file for
//  ease of distribution and compilation.  Other than that,
//  the code is essentially unchanged.
//
//  This program takes two arguments: an input file and an output
//  file.  You can leave off one argument and send your output to
//  stdout.  Leave off two arguments and read your input from stdin
//  as well.
//
//

#if !defined( _ARI_H )
#define _ARI_H

#include "st.h"

#define No_of_chars 256                 /* Number of character symbols      */
#define EOF_symbol (No_of_chars+1)      /* Index of EOF symbol              */
#define No_of_symbols (No_of_chars+1)   /* Total number of symbols          */

/* TRANSLATION TABLES BETWEEN CHARACTERS AND SYMBOL INDEXES. */

int char_to_index[No_of_chars];         /* To index from character          */
unsigned char index_to_char[No_of_symbols+1]; /* To character from index    */

/* ADAPTIVE SOURCE MODEL */

int freq[No_of_symbols+1];      /* Symbol frequencies                       */
int cum_freq[No_of_symbols+1];  /* Cumulative symbol frequencies            */

/* DECLARATIONS USED FOR ARITHMETIC ENCODING AND DECODING */

/* SIZE OF ARITHMETIC CODE VALUES. */

#define Code_value_bits 16              /* Number of bits in a code value   */
typedef long code_value;                /* Type of an arithmetic code value */
#define Top_value (((long)1<<Code_value_bits)-1)      /* Largest code value */

/* HALF AND QUARTER POINTS IN THE CODE VALUE RANGE. */

#define First_qtr (Top_value/4+1)       /* Point after first quarter        */
#define Half	  (2*First_qtr)         /* Point after first half           */
#define Third_qtr (3*First_qtr)         /* Point after third quarter        */


/* CUMULATIVE FREQUENCY TABLE. */

#define Max_frequency 16384-1           /* Maximum allowed frequency count  */
extern int cum_freq[No_of_symbols+1];   /* Cumulative symbol frequencies    */

void start_model(void);
void start_encoding(void);
void start_decoding(char_vector_iterator& i);
void encode_symbol(char_vector& code, int symbol,int cum_freq[]);
int decode_symbol(char_vector_iterator& i, int cum_freq[]);
void update_model(int symbol);
void done_encoding(char_vector& code);
void done_outputing_bits(char_vector& code);

/* CURRENT STATE OF THE ENCODING. */

code_value low, high;           /* Ends of the current code region          */
long bits_to_follow;            /* Number of opposite bits to output after  */
                                /* the next bit.                            */
/* CURRENT STATE OF THE DECODING. */

static code_value value;        /* Currently-seen code value                */
//static code_value low, high;    /* Ends of current code region              */

/* BIT OUTPUT ROUTINES. */

/* THE BIT BUFFER. */

int buffer;                     /* Bits buffered for output                 */
int bits_to_go;                 /* Number of bits free in buffer            */
int garbage_bits;               /* Number of bits past end-of-file          */

/* OUTPUT A BIT. */

inline void output_bit(char_vector& code, int bit) { 
    buffer >>= 1;
    if (bit) buffer |= 0x80;                    /* Put bit in top of buffer.*/
    bits_to_go -= 1;
    if (bits_to_go==0) {                        /* Output buffer if it is   */
        code.push_back((char)buffer);           /* now full.                */
        bits_to_go = 8;
    }
}

/* OUTPUT BITS PLUS FOLLOWING OPPOSITE BITS. */

void bit_plus_follow(char_vector& code, int bit) {
  output_bit(code, bit);                      /* Output the bit.          */
  while (bits_to_follow>0) {
    output_bit(code, !bit);                   /* Output bits_to_follow    */
    bits_to_follow -= 1;                      /* opposite bits. Set       */
  }                                           /* bits_to_follow to zero.  */
}

/* INPUT A BIT. */

inline int input_bit(char_vector_iterator& i) {
  int t;
  if (bits_to_go == 0) {                       /* Read the next byte if no */
    buffer = *i;                              /* bits are left in buffer. */
    ++i;
    if (buffer == EOF) {
      garbage_bits += 1;                      /* Return arbitrary bits */
      //if (garbage_bits > Code_value_bits - 2) { /* after eof, but check  */
      //fprintf(stderr,"Bad input file\n");     /* for too many such.    */
      //exit(-1);
      //}
    }
    bits_to_go = 8;
  }
  t = buffer&1;                               /* Return the next bit from */
  buffer >>= 1;                               /* the bottom of the byte.  */
  bits_to_go -= 1;
  return t;
}

//
// main function for the encoder
//
char_vector arith_encoder(char_vector& text) {
  char_vector
    code;
  int
    symbol;

  start_model();                              /* Set up other modules.    */
  buffer = 0;                                 /* Buffer is empty to start */
  bits_to_go = 8;                             /* with.                    */
  start_encoding();
  for (char_vector_iterator i = text.begin(); i !=text.end(); ++i) { /* Loop through characters. */
    symbol = char_to_index[(unsigned short)(*i & 0xFF)]; /* Translate to an index.   */
    encode_symbol(code, symbol,cum_freq);     /* Encode that symbol.      */
    update_model(symbol);                     /* Update the model.        */
  }
  encode_symbol(code,EOF_symbol,cum_freq);    /* Encode the EOF symbol.   */
  done_encoding(code);                        /* Send the last few bits.  */
  code.push_back((char)(buffer >> bits_to_go));
  return code;
}

//
// main function for the decoder
//
char_vector arith_decoder(char_vector& code) {
  char_vector
    text;
  char_vector_iterator
    i;
  int
    symbol;

  start_model();                              /* Set up other modules.    */
  bits_to_go = 0;                             /* Buffer starts out with   */
  garbage_bits = 0;                           /* no bits in it.           */
  i = code.begin();
  start_decoding(i);
  for (;;) {                                  /* Loop through characters. */
    symbol = decode_symbol(i, cum_freq);      /* Decode next symbol.      */
    if (symbol == EOF_symbol) break;          /* Exit loop if EOF symbol. */
    text.push_back((char)index_to_char[symbol]);  /* Translate to a character.*/
    update_model(symbol);                         /* Update the model.        */
  }
  return text;
}

/* -------------------  ARITHMETIC ENCODING ALGORITHM ------------------  */


/* START ENCODING A STREAM OF SYMBOLS. */

void start_encoding(void) {
  low = 0;                                    /* Full code range.         */
  high = Top_value;
  bits_to_follow = 0;                         /* No bits to follow next.  */
}


/* ENCODE A SYMBOL. */

void encode_symbol(char_vector& code, int symbol,int cum_freq[]) {
  long range;                          /* Size of the current code region */
  range = (long)(high-low)+1;
  high = low +                                /* Narrow the code region   */
    (range*cum_freq[symbol-1])/cum_freq[0]-1; /* to that allotted to this */
  low = low +                                 /* symbol.                  */
    (range*cum_freq[symbol])/cum_freq[0];
  for (;;) {                                  /* Loop to output bits.     */
    if (high<Half) {
      bit_plus_follow(code, 0);               /* Output 0 if in low half. */
    } else if (low>=Half) {                   /* Output 1 if in high half.*/
      bit_plus_follow(code, 1);
      low -= Half;
      high -= Half;                           /* Subtract offset to top.  */
    } else if (low>=First_qtr                 /* Output an opposite bit   */
	       && high<Third_qtr) {           /* later if in middle half. */
      bits_to_follow += 1;
      low -= First_qtr;                       /* Subtract offset to middle*/
      high -= First_qtr;
    } else break;                             /* Otherwise exit loop.     */
    low = 2*low;
    high = 2*high+1;                          /* Scale up code range.     */
  }
}


/* FINISH ENCODING THE STREAM. */

void done_encoding(char_vector& code) {
  bits_to_follow += 1;                        /* Output two bits that     */
  if (low<First_qtr)
    bit_plus_follow(code, 0);                 /* select the quarter that  */
  else
    bit_plus_follow(code, 1);                 /* the current code range   */
}                                               /* contains.                */


/* START DECODING A STREAM OF SYMBOLS. */

void start_decoding(char_vector_iterator& ci) {
  int i;
  value = 0;                                  /* Input bits to fill the   */
  for (i = 1; i<=Code_value_bits; i++) {      /* code value.              */
    value = 2*value+input_bit(ci);
  }
  low = 0;                                    /* Full code range.         */
  high = Top_value;
}


/* DECODE THE NEXT SYMBOL. */

int decode_symbol(char_vector_iterator& i, int cum_freq[]) {
  long range;                 /* Size of current code region              */
  int cum;                    /* Cumulative frequency calculated          */
  int symbol;                 /* Symbol decoded                           */
  range = (long)(high-low)+1;
  cum = (int)                                 /* Find cum freq for value. */
    ((((long)(value-low)+1)*cum_freq[0]-1)/range);
  for (symbol = 1; cum_freq[symbol]>cum; symbol++) ; /* Then find symbol. */
  high = low +                                /* Narrow the code region   */
    (range*cum_freq[symbol-1])/cum_freq[0]-1; /* to that allotted to this */
  low = low +                                 /* symbol.                  */
    (range*cum_freq[symbol])/cum_freq[0];
  for (;;) {                                  /* Loop to get rid of bits. */
    if (high<Half) {
      /* nothing */                           /* Expand low half.         */
    } else if (low>=Half) {                   /* Expand high half.        */
      value -= Half;
      low -= Half;                            /* Subtract offset to top.  */
      high -= Half;
    } else if (low>=First_qtr                 /* Expand middle half.      */
	     && high<Third_qtr) {
      value -= First_qtr;
      low -= First_qtr;                       /* Subtract offset to middle*/
      high -= First_qtr;
    } else break;                             /* Otherwise exit loop.     */
    low = 2*low;
    high = 2*high+1;                          /* Scale up code range.     */
    value = 2*value+input_bit(i);             /* Move in next input bit.  */
  }
  return symbol;
}

/* THE ADAPTIVE SOURCE MODEL */

/* INITIALIZE THE MODEL. */

void start_model(void) {
  int i;
  for (i = 0; i<No_of_chars; i++) {           /* Set up tables that       */
    char_to_index[i] = i+1;                   /* translate between symbol */
    index_to_char[i+1] = (unsigned char) i;   /* indexes and characters.  */
  }
  for (i = 0; i<=No_of_symbols; i++) {        /* Set up initial frequency */
    freq[i] = 1;                              /* counts to be one for all */
    cum_freq[i] = No_of_symbols-i;            /* symbols.                 */
  }
  freq[0] = 0;                                /* Freq[0] must not be the  */
}                                             /* same as freq[1].         */


/* UPDATE THE MODEL TO ACCOUNT FOR A NEW SYMBOL. */

void update_model(int symbol) {
  int i;                                      /* New index for symbol     */
  if (cum_freq[0]==Max_frequency) {           /* See if frequency counts  */
    int cum;                                  /* are at their maximum.    */
    cum = 0;
    for (i = No_of_symbols; i>=0; i--) {      /* If so, halve all the     */
      freq[i] = (freq[i]+1)/2;                /* counts (keeping them     */
      cum_freq[i] = cum;                      /* non-zero).               */
      cum += freq[i];
    }
  }
  for (i = symbol; freq[i]==freq[i-1]; i--) ; /* Find symbol's new index. */
  if (i<symbol) {
    int ch_i, ch_symbol;
    ch_i = index_to_char[i];                  /* Update the translation   */
    ch_symbol = index_to_char[symbol];        /* tables if the symbol has */
    index_to_char[i] = (unsigned char) ch_symbol;     /* moved.           */
    index_to_char[symbol] = (unsigned char) ch_i;
    char_to_index[ch_i] = symbol;
    char_to_index[ch_symbol] = i;
  }
  freq[i] += 1;                               /* Increment the frequency  */
  while (i>0) {                               /* count for the symbol and */
    i -= 1;                                   /* update the cumulative    */
    cum_freq[i] += 1;                         /* frequencies.             */
  }
}

#endif

/*
#include <fstream.h>

void print_vec_char (char_vector& x) {
  for (char_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << *i;
}

int main(int argc, char* argv[]) {
  ifstream
    ifs(argv[1]);
  
  if (!ifs) {
    cout << "Couldn't open the input file" << endl;
    exit(0);
  }
  
  char_vector b;
  
  for (unsigned char ch;;) {
    ifs.get(ch);
    if (ifs)
      b.push_back(ch);
    else
      break;
  }
  
  char_vector
    a = arith_encoder(b);

  cerr << b.size() << " " << a.size() << endl;

  char_vector
    c = arith_decoder(a);

  print_vec_char(c);
}
*/
