//
// Suffix Trees Construction
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, Stefano Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 28.10.98 $
//

// #define DEBUGSEARCH
// #define DEBUGTREE
// #define DEBUGSTAT
#include "st.h"

extern text_vector dict_vec;
extern void print_vec_char (char_vector&);
extern void print_text_vec (text_vector&);
extern double hTEXT, hTXT, hOCC, hLEN, hPOS;
extern base2 loga;

#ifdef OFF3
#ifdef OFF2
inline double gain(int f, int m, int n) { // OFF2
  //
  // gain function
  //
  return (f - 1) * (hTEXT * m - loga.logup(n) - loga.logup(m));
}
#else
inline double gain(int f, int m, int n) { // OFF3
  //
  // gain function
  //
  return (f - 1) * hTEXT * m - f * loga.logup(n) - loga.logup(m);
}
#endif
#else
inline double gain(int f, int m, int n) { // OFF1
  //
  // gain function
  //
  return (f - 1) * hTEXT * m - loga.logup(m) - loga.logup(f) - f * loga.logup(n);
  /* if (hTXT == 0.0)
     g = (f - 1) * hTEXT * m - log2(m) - log2(f) - f * log2(n);
     else
     g = f * hTEXT * m - hTXT * m - hLEN - hOCC - f * hPOS; */
}
#endif

void Tree::print(const text_vector& t, int l = 1) {
  //
  // Print the tree
  //
  Tree
    *c = get_child(),
    *s = get_sibling();

  for (int i = 1; i != l; ++i)
    cout << " ";

  for (text_vector_iterator t_i = this->begin; t_i != this->end; ++t_i)
    cout << *t_i;
  cout << " [" << (get_begin() - t.begin()) << ",";
  cout << (get_end() - t.begin()) << ",";
  cout << get_count() << "]" << endl;

  if (c)
    c->print(t, l + 1);

  if (s)
    s->print(t, l); 
}


int Tree::count_nodes(void) {
  //
  // count the number of nodes in the tree
  //
  Tree
    *c = get_child(),
    *s = get_sibling();
  int
    tmp = 0;

  if (s)
    tmp = s->count_nodes(); 

  if (c)
    tmp += c->count_nodes();

  return (tmp + 1);
}


void Tree::insert_suffix(Tree& a, int l) {
  //
  // insert the node a (a suffix) in the tree 
  //

  //# ifdef PRUNED
  //if (l >= MAX_LENGTH)
  //  return;
  //# endif

  Tree
    *p;
  text_vector_iterator
    i = get_begin(),
    j = get_end(),
    k = a.get_begin();
  pair_text_vector_iterator
    r = mismatch(i,j,k); // is tx[t->b..t->e] a substring of tx[a->b..a->e] ??
  int
    matched_length = r.first - i;
  
  if (r.first == i) {
#   ifdef DEBUGTREE
    cout << "No match" << endl;
    cout << matched_length << " chars matched" << endl;
#   endif
    p = get_sibling();
    if (p)
      p->insert_suffix(a, l);
    //#   ifdef PRUNED
    //else if (MAX_LENGTH - l > a.get_length()) 
    //  insert_sibling(&a);
    //else {
    //  a.set_end(MAX_LENGTH - l + a.get_begin());
    //  insert_sibling(&a);
    //}
    //#   else
    else
      insert_sibling(&a);
    //#   endif
  } else if (r.first == j) {
#   ifdef DEBUGTREE
    cout << "Full match of " << k << " with ";
    cout << i << " first mismatch " << *(r.second) << endl;
    cout << matched_length << " chars matched" << endl;
#   endif

    a.set_begin(r.second);
    increment_count();
    p = get_child();
    if (p)
      p->insert_suffix(a, l + matched_length);
    else
      insert_sibling(&a);
  } else {
#   ifdef DEBUGTREE
    cout << "Partial match of " << k << " with ";
    cout << i << " first mismatch " << *(r.second) << endl;
    cout << matched_length << " chars matched" << endl;
#   endif

    p = new Tree(r.first, j, get_count());
    set_end(r.first);
    increment_count();
    insert_child(p);
    if (r.second < a.get_end()) {
	a.set_begin(r.second);
	get_child()->insert_suffix(a, l + matched_length);
    }
  }
}


void Tree::insert_suffix_noover(Tree& a, const text_vector& tx, int l) {
  //
  // insert the node a (a suffix) in this Tree (nonverlapping version)
  // (tx is not necessary - only for debug)
  //
  //# ifdef PRUNED
  //if (l >= MAX_LENGTH)
  //  return;
  //# endif

  text_vector_iterator
    t_b = get_begin(),
    t_e = get_end(),
    a_b = a.get_begin(),
#   ifdef DEBUGTREE
    a_e = a.get_end(),
#   endif
    over_patt_mismatch,
    brek;
  pair_text_vector_iterator
    r = mismatch(t_b, t_e, a_b);
  text_vector_iterator
    text_mismatch = r.first,
    patt_mismatch = r.second;
  int
    over_matched_length,
    matched_length = text_mismatch - t_b;

# ifdef DEBUGTREE
  cout << "length=" << l << " matched=" << matched_length << endl;
# endif

  if (matched_length == 0) {
#   ifdef DEBUGTREE
    cout << "No match" << endl << endl;
#   endif
    Tree
      *p = get_sibling();
    if (p)
      p->insert_suffix_noover(a, tx, l);
    //#   ifdef PRUNED
    //else if (MAX_LENGTH - l > a.get_length()) 
    //  insert_sibling(&a);
    //else {
    //  a.set_end(MAX_LENGTH - l + a.get_begin());
    //  insert_sibling(&a);
    // }
    //#   else
    else 
      insert_sibling(&a);
    //#   endif
  } else if (a_b - l < text_mismatch) { // they match and there is overlap
    brek = a_b - l;
    over_matched_length = brek - t_b;
    over_patt_mismatch = a_b + over_matched_length;
    if (over_matched_length > 0) { // part of the match in non overlapped
#     ifdef DEBUGTREE
      cout << "Partial overlapped match of pattern ";
      cout << a << "(" << (a_b - tx.begin()) << ",";
      cout << (a_e - tx.begin()) << ") " << endl << " with text ";
      cout << this << "(" << (t_b - tx.begin()) << ",";
      cout << (t_e - tx.begin()) << ") first mismatch (pat,txt)=";
      cout << "(" << (patt_mismatch - tx.begin()) << ",";
      cout << (text_mismatch - tx.begin()) << ")" << endl;
      cout << " Actually matched: (" << (a_b - tx.begin());
      cout << "," << (over_patt_mismatch - tx.begin()) << ") l=";
      cout << over_matched_length << endl << endl;
#     endif
      insert_child(new Tree(brek, t_e, get_count()));
      update_node(a_b, over_patt_mismatch);
      a.set_begin(over_patt_mismatch);
      get_child()->insert_suffix_noover(a, tx, l + over_matched_length);
    } else { // all the match is overlapped (no frequency increment occurrs)
      a.set_begin(patt_mismatch);
      if (text_mismatch == t_e) {
#       ifdef DEBUGTREE
	cout << "Overlapped full match of pattern ";
	cout << a << "(" << (a_b - tx.begin()) << ",";
	cout << (a_e - tx.begin()) << ") " << endl << " with text ";
	cout << this << "(" << (t_b - tx.begin()) << ",";
	cout << (t_e - tx.begin()) << ") first mismatch (pat,txt)=";
	cout << "(" << (patt_mismatch - tx.begin()) << ",";
	cout << (text_mismatch - tx.begin()) << ")" << endl;
#       endif
	//get_child can be NULL due to lacking $
	if (a.get_begin() != a.get_end()) {
	  Tree
	    *c = get_child();
	  if (c)
	    c->insert_suffix_noover(a, tx, l + matched_length);
	  else
	    insert_child(&a);
	}
      } else { 
#       ifdef DEBUGTREE
	cout << "Overlapped partial match of pattern ";
	cout << a << "(" << (a_b - tx.begin()) << ",";
	cout << (a_e - tx.begin()) << ") " << endl << " with text ";
	cout << this << "(" << (t_b - tx.begin()) << ",";
	cout << (t_e - tx.begin()) << ") first mismatch (pat,txt)=";
	cout << "(" << (patt_mismatch - tx.begin()) << ",";
	cout << (text_mismatch - tx.begin()) << ")" << endl;
#       endif
	set_end(text_mismatch); // remain the old occurrence!
	insert_child(new Tree(text_mismatch, t_e, get_count()));
	if (a.get_begin() != a.get_end())
	  get_child()->insert_suffix_noover(a, tx, l + matched_length);
      }
    }
  } else if (text_mismatch == t_e) {
#   ifdef DEBUGTREE
    cout << "Non overlapped full match of pattern ";
    cout << a << "(" << (a_b - tx.begin()) << ",";
    cout << (a_e - tx.begin()) << ")" << endl << " with text ";
    cout << this << "(" << (t_b - tx.begin()) << ",";
    cout << (t_e - tx.begin()) << ") first mismatch (pat,txt)=";
    cout << "(" << (patt_mismatch - tx.begin()) << ",";
    cout << (text_mismatch - tx.begin()) << ")" << endl << endl;
#   endif
    update_node(a_b, patt_mismatch);
    a.set_begin(patt_mismatch);
    //get_child can be NULL due to lacking $
    if (a.get_begin() != a.get_end()) {
      Tree
	*c = get_child();
      if (c)
	c->insert_suffix_noover(a, tx, l + matched_length);
      else
	insert_child(&a);
    }
  }
  else { // text_mismatch + l <= a_b
#   ifdef DEBUGTREE
    cout << "Non overlapped partial match of pattern ";
    cout << a << "(" << (a_b - tx.begin()) << ",";
    cout << (a_e - tx.begin()) << ")" << endl << " with text ";
    cout << this << "(" << (t_b - tx.begin()) << ",";
    cout << (t_e - tx.begin()) << ") first mismatch (pat,txt)=";
    cout << "(" << (patt_mismatch - tx.begin()) << ",";
    cout << (text_mismatch - tx.begin()) << ")" << endl << endl;
#   endif
    a.set_begin(patt_mismatch);
    insert_child(new Tree(text_mismatch, t_e, get_count()));
    update_node(a_b, patt_mismatch);
    if (a.get_begin() != a.get_end())
      get_child()->insert_suffix_noover(a, tx, l + matched_length);
  }
}



Tree* Tree::search(const text_vector& tx, text_vector& vs) {
  //
  // search a string in the node
  //
  Tree
    *p;
  text_vector_iterator
    i = get_begin(),
    j = get_end(),
    k = vs.begin();
  pair_text_vector_iterator
    r = mismatch(i,j,k);
  
# ifdef DEBUGSEARCH
  cout << "DebugSearch: ";
  print_vec_char(vs);
  cout << endl;
# endif

  if (r.first == i) {
#   ifdef DEBUGSEARCH
    cout << "No match" << endl;
#   endif
    p = get_sibling();
  }
  else {
#   ifdef DEBUGSEARCH
    cout << "Partial match of " << k << " with ";
    cout << i << " first mismatch " << *r.second << endl;
#   endif
    vs.erase(vs.begin(), r.second);
    p = get_child();
  }
  if (vs.empty())
    return this;
  else if (p)
    return p->search(tx, vs);
  else 
    return (Tree*)NULL;
}



void Tree::build_suffix_tree(text_vector& tx) {
  Tree
    *p;
  for (text_vector_iterator t_i = tx.begin() + 1; t_i != tx.end(); ++t_i) {
#   ifdef PRUNED
    p = new Tree(t_i,  min(tx.end(), t_i + MAX_LENGTH));
#   else
    p = new Tree(t_i, tx.end());
#   endif

#   ifdef DEBUGTREE
    cout << "------------------------" << endl;
    cout << "Processing(" << (t_i - tx.begin()) << ") ";
    cout << p << endl;
#   endif

    insert_suffix(*p, 0);

#   ifdef DEBUGTREE
    cout << "Tree = " << endl;
    print(tx); 
    cout << endl;
#   endif
  }
}



void Tree::build_augmented_suffix_tree(text_vector_iterator start, text_vector& tx) {
  //
  // creates the augmented suffix tree
  //
  Tree
    *p;
  for (text_vector_iterator t_i = start; t_i != tx.end(); ++t_i) {
#   ifdef NO_RECURSIVE_PTR
    text_vector_iterator
      t_j;
    if (t_i->second == POINTER)
      continue;
    for (t_j = t_i; t_j != tx.end(); ++t_j)
      if (t_j->second == POINTER)
	break;
    // cout << t_j - t_i << " " << tx.end() - t_i << endl;
#   ifdef PRUNED
    p = new Tree(t_i,  min(t_j, t_i + MAX_LENGTH));
#   else
    p = new Tree(t_i, t_j);
#   endif

#   else

#   ifdef PRUNED
    p = new Tree(t_i,  min(tx.end(), t_i + MAX_LENGTH));
#   else
    p = new Tree(t_i, tx.end());
#   endif
#   endif

#   ifdef DEBUGTREE
    cout << "------------------------" << endl;
    cout << "Processing(" << t_i - tx.begin() << "," << t_j - tx.begin() << ") ";
    cout << p << endl;
#   endif
    
    insert_suffix_noover(*p, tx, 0);

#   ifdef DEBUGTREE
    cout << "Tree = " << endl;
    print(tx); 
    cout << endl;
#   endif
  }
}



void Tree::compute_gain(stat_queue& sq, const text_vector& txt, int prefix_m) {
  //
  // Compute the gain function on each node
  //
  static int
    n = txt.size();
  int
    substring_m = get_end() - get_begin(),
    m = prefix_m + substring_m,
    f = get_count();
  text_vector
    s(get_end() - m, get_end());
  Tree
    *chi = get_child(),
    *sib = get_sibling();
  
# ifdef DEBUGSTAT
  print_text_vec(s);
  cout << " " << substring_m << " " << m << " " << f << " " << n << endl;
# endif
  
  if ((f >= MIN_OCC) && (m >= MIN_LENGTH)) {
    double
      g = gain(f,m,n);

#   ifdef DEBUGSTAT
    print_text_vec(s);
    cout << " : " << g << " bits, freq " << f << endl;
#   endif

    if (sq.empty()) {
      sq.push(Stat(g, get_begin(), this, m, f, s));
#     ifdef DEBUGSTAT
      cout << " empty pushed" << endl;
#     endif
    } else {
      Stat
	stat_top(sq.top());
      text_vector_iterator
	bq = stat_top.get_string_begin(),
	eq = stat_top.get_string_end();
      if ((std::search(bq, eq, s.begin(), s.end()) == eq) &&
	  (std::search(s.begin(), s.end(), bq, eq)) == s.end()) {
	sq.push(Stat(g, get_begin(), this, m, f, s));
#       ifdef DEBUGSTAT
	cout << " not a substring" << endl;
#       endif
      } else if (stat_top.get_gain() < g) { // one is a substring of the other
	sq.pop();
	sq.push(Stat(g, get_begin(), this, m, f, s));      
#       ifdef DEBUGSTAT
	cout << " exchanged with the top ";
	print_text_vec(stat_top.get_string());
	cout << endl;
	sq.print_queue();
#       endif
      }
    }
  }

  if (chi)
    chi->compute_gain(sq, txt, m);

  if (sib) 
    sib->compute_gain(sq, txt, prefix_m);
}



void Tree::clear_terminators(void) {
  //
  // remove the dollars from the tree
  //
  Tree
    *c = get_child(),
    *s = get_sibling();

  if (*(get_end() - 1) == TERMINATOR)
    set_end(get_end() - 1);

  if (c)
    if (*(c->get_begin()) == TERMINATOR) {
      Tree
	*ps = c->get_sibling();
      set_child(ps);
      delete(c);
      if (ps)
	ps->clear_terminators();
   } else
     c->clear_terminators();

  if (s)
    if (*(s->get_begin()) == TERMINATOR) {
      Tree
	*ps = s->get_sibling();
      set_sibling(ps);
      delete(s);
      if (ps)
	ps->clear_terminators();
   } else
      s->clear_terminators(); 
}
