//
// Suffix Trees Construction
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 10.28.98 $
//

#if !defined( _ST_H )
#define _ST_H

#include <string>
#include <vector.h>
#include <stack.h>
#include <list.h>
#include <algo.h>
#include <math.h>

#define OFF2                  // is a special case of OFF3
#define OFF3
#define NO_RECURSIVE_PTR      // disable together with OFF3
//#define RECOMPUTE_ENCODING 
#define PRUNED
#define TERMINATOR (char)255
#define THRE       0.0        // when stop the iterative substitution? gain<THRE?
#define QUEUE_LEN  10         // the length of the queue
#define MAX_LENGTH 800        // consider only matches <= MAX_LENGTH (turn on PRUNED)
#define MIN_OCC    2          // consider only matches whose occurrences >= MIN_OCC
#define MIN_LENGTH 2          // consider only matches >= MIN_LENGTH
#define MAXSHORT   65535
#define MAXCHAR    255
#define B          8.0        // literal size in bits
#define TRUE       1
#define FALSE	   0
#define LITERAL    1
#define POINTER    0

typedef vector<char> char_vector;
typedef char_vector::iterator char_vector_iterator;
typedef ostream_iterator<char> char_ostream;
typedef istream_iterator<char, ptrdiff_t> char_input;
class Stat;
class Tree;
#ifdef OFF3
class symbol {
public:
  char first;
  bool second;

  symbol() : first(char()), second(bool()) {}
  symbol(const symbol& a) : first(a.first), second(a.second) {}
  symbol(const char& a) : first(a), second(LITERAL) {}
  symbol(const char& a, const bool& b) : first(a), second(b) {}

  friend ostream& operator<< (ostream& st, symbol x) {
    if (x.second)
      st << x.first;
    else
      st << "@" << (unsigned short)(x.first & 0xFF);
    return st;
  }

  symbol& operator=(const symbol& x) {
    first = x.first;
    second = x.second;
    return *this;
  }

  friend bool operator==(const symbol& x, const symbol& y) { 
    return (x.first == y.first && x.second == y.second);
  }

  friend bool operator<(symbol& x, const symbol& y) { 
    return x.first < y.first || (!(y.first < x.first) && x.second < y.second); 
  }
};
#else
typedef char symbol;
#endif

typedef vector<symbol> text_vector;
typedef text_vector::iterator text_vector_iterator;
typedef pair<text_vector_iterator,text_vector_iterator> pair_text_vector_iterator;

template <class Container, class Compare> 
class prioqueue {
public:
  typedef typename Container::value_type value_type;
  typedef typename Container::size_type size_type;
  typedef typename Container::iterator iterator_type;
  typedef typename Container::const_reference const_reference;
protected:
  Container c;
  Compare comp;
  int maxsize; // max size of the queue
public:
  prioqueue(const Compare& x = Compare()) :  c(), comp(x) {}
  prioqueue(const int y, const Compare& x = Compare()) :
    c(), comp(x), maxsize(y) {}
  bool empty() const { return c.empty(); }
  size_type size() const { return c.size(); }
  //value_type& top() {
  //  return c.back(); //c.front();
  // }
  const_reference top() const {
    return c.back(); //c.front();
  }
  void push(const value_type& x) { 
    iterator_type i = c.begin();
    while ((comp(x,*i)) && (i != c.end()))
      ++i;
    c.insert(i, x); // better if the container is a list
    if ((int)c.size() > maxsize)
      c.pop_front();
  }
  void pop() {
    c.erase(--c.end());
  }
  void print_queue () {
    cout << "Queue size: " << c.size() << " maxsize: " << maxsize << endl;
    for (iterator_type i = c.begin(); i !=c.end(); ++i)
      cout << *i << endl;
  }
};



class Stat {
  double
    gain;                // gain
  text_vector_iterator
    txt_ptr;             // pointer
  Tree
    *tree_ptr;           // pointer in the Tree
  int
    len,                 // length of the substring
    freq;                // the # of occurrences
  text_vector
    str;                 // the string itself  

public:
  // ----- Constructors
  Stat() :
    gain(0.0), txt_ptr(0), tree_ptr(NULL), len(0), freq(0), str() {}

  Stat(const double g) :
    gain(g), txt_ptr(0), tree_ptr(NULL), len(0), freq(0), str() {}

  Stat(const double g, const text_vector_iterator t, Tree *tr,
       const int l, const int f, const text_vector& s) :
    gain(g), txt_ptr(t), tree_ptr(tr), len(l), freq(f), str(s) {}

  bool operator<(const Stat &s) const {
    return gain < s.gain;
  }

  friend ostream& operator<< (ostream& st, Stat s) {
    st << "String= ";
    for (text_vector_iterator i = s.str.begin(); i != s.str.end(); ++i)
      st << *i;
    st << " Gain= " << s.gain;
    st << " Frequency= " << s.freq;
    // st << " Text pointer= " << s.txt_ptr;
    // st << " Tree pointer= " << s.tree_ptr;
    st << " Length= " << s.len << endl;
    return st;
  }

  // ----- GET accessors
  double get_gain(void) { return gain; }
  text_vector_iterator get_txt_ptr(void) { return txt_ptr; }
  Tree* get_tree_ptr(void) { return tree_ptr; }
  int get_len(void) { return len; }
  int get_freq(void) { return freq; }
  text_vector& get_string(void) { return str; }
  text_vector_iterator get_string_begin(void) { return str.begin(); }
  text_vector_iterator get_string_end(void) { return str.end(); }

  // ----- SET accessors
  Stat* set_gain(double x) { gain = x; return this; }
  Stat* set_txt_ptr(text_vector_iterator i) { txt_ptr = i; return this; }
  Stat* set_tree_ptr(Tree* p) { tree_ptr = p; return this; }
  Stat* set_len(int x) { len = x; return this; }
  Stat* set_freq(int x) { freq = x; return this; }
  Stat* set_string(text_vector& s) { str = s; return this; }
  Stat* init() { gain = 0.0; txt_ptr = 0; tree_ptr = 0; len = 0; freq = 0;
  str.erase(str.begin(), str.end()); return this; }
};


//template <class Container, class Compare> class prioqueue;
typedef prioqueue<list<Stat>, greater<Stat> > stat_queue;

class Tree {
  text_vector_iterator
    begin,             // begin pointer in text
    end;               // end pointer in text (+1)
  Tree
    *sibling,          // sibling pointer in the Tree
    *child;            // child pointer in the Tree
  int
    count;             // frequency of occurrency
    
public:
  // ----- Constructors
  Tree() :
    begin(0), end(0), sibling(NULL), child(NULL), count(0) {}
  Tree(text_vector_iterator ss, text_vector_iterator ee) :
    begin(ss), end(ee), sibling(NULL), child(NULL), count(1) {}
  Tree(text_vector_iterator ss, text_vector_iterator ee, int cc) :
    begin(ss), end(ee), sibling(NULL), child(NULL), count(cc) {}

  // ----- Destructors
  ~Tree() {
    Tree
      *c = get_child(),
      *s = get_sibling();

    if (c) {
      Tree
	*cc = c->get_child(),
	*cs = c->get_sibling();
      if ((cc)||(cs)) {
	c->~Tree();
	this->~Tree();
      } else {
	delete c;
	set_child(NULL);
      }
    }
    if (s) {
      Tree
	*sc = s->get_child(),
	*ss = s->get_sibling();
      if ((sc)||(ss)) {
	s->~Tree();
	this->~Tree();
      } else {
	delete s;
	set_sibling(NULL);
      }
    }
  }

  // ----- GET accessors
  text_vector_iterator get_begin(void) { return begin; }
  text_vector_iterator get_end(void) { return end; }
  Tree* get_sibling(void) { return sibling; }
  Tree* get_child(void) { return child; }
  int get_count(void) { return count; }
  int get_length(void) { return end - begin; }

  // ----- SET accessors
  void set_begin(text_vector_iterator i) { begin = i; }
  void set_end(text_vector_iterator i) { end = i; }
  Tree* set_sibling(Tree* n) { sibling = n; return this; }
  Tree* set_child(Tree* n) { child = n; return this; }
  Tree* set_count(int i) { count = i; return this; }
  Tree* increment_count(void) { count += 1; return this; }
  Tree* update_node(text_vector_iterator b, text_vector_iterator e) { begin = b; end = e; count += 1; return this; }

  // ----- Operators
  friend ostream& operator<< (ostream& st, Tree a) {
    for (text_vector_iterator t_i = a.begin; t_i != a.end; ++t_i)
      st << *t_i;
    return st;
  }

  // ----- Other methods
  void insert_child(Tree* b) {
    Tree*
      c = child;
    child = b;
    b->child = c;
  }

  void insert_sibling(Tree* b) {
    Tree*
      c = sibling;
    sibling = b;
    b->sibling = c;
  }

  void insert_suffix(Tree&, int);
  void insert_suffix_noover(Tree&, const text_vector&, int);
  void print(const text_vector&, int);
  int count_nodes(void);
  Tree* search(const text_vector&, text_vector&);
  void build_suffix_tree(text_vector&);
  void build_augmented_suffix_tree(text_vector_iterator, text_vector&);
  void compute_gain(stat_queue&, const text_vector&, int);
  void clear_terminators(void);
}; 


inline double log2(int x) {
  //
  // return the code lenght for x
  //
  if (x > 0)
    return ceil(log(x)/log(2));
  else
    return 1;
}

struct base2 {

int log_down(int num) {
  switch (sizeof(int)) {
  case 4:
    if (num&0xffff0000) {
      if (num&0xff000000) {
	if (num&0xf0000000) {
	  if (num&0xc0000000) {
	    if (num&0x80000000)
	      return 31;
	    return 30; }
	  if (num&0x20000000)
	    return 29;
	  return 28; }
	if (num&0x0c000000) {
	  if (num&0x08000000)
	    return 27; 
	  return 26; }
	if (num&0x02000000)
	  return 25;
	return 24; }
      if (num&0x00f00000) {
	if (num&0x00c00000) {
	  if (num&0x00800000)
	    return 23;
	  return 22; }
	if (num&0x00200000)
	  return 21;
	return 20; }
      if (num&0x000c0000)
	{ if (num&0x00080000)
	  return 19;
	return 18; }
      if (num&0x00020000)
	return 17;
      return 16; }
    if (num&0xff00) {
      if (num&0xf000) {
	if (num&0xc000) {
	  if (num&0x8000)
	    return 15;
	  return 14; }
	if (num&0x2000)
	  return 13;
	return 12; }
      if (num&0x0c00) {
	if (num&0x0800)
	  return 11;
	return 10; }
      if (num&0x0200)
	return 9;
      return 8; }
    if (num&0xf0) {
      if (num&0xc0) {
	if (num&0x80)
	  return 7;
	return 6;
      } if (num&0x20)
	return 5;
      return 4; }
    if (num&0xc) {
      if (num&0x08)
	return 3;
      return 2; }
    if (num&0x2)
      { return 1; }
    if (num&0x1)
      return 0;
    return -1;
  }
  
  const int
    s_upperbit = 1 << (8 * sizeof(int) - 1);
  int
    lg = 8 * sizeof(int) - 1; // assumed: lg & upper = 0
  while (! ((num & s_upperbit) | (lg & s_upperbit)))
    num <<= 1, lg--;
  return lg;
}

int logdown(int num) {
  //
  // return log to base 2 (rounded down); logdown(0)=-1
  //
  int
    lg = log_down(num);
  return lg;
}

int logup(int num) {
  //
  // return log to base 2 (rounded up)  ; logup  (0)=-1
  //
  int
    lg = log_down(num);
  return num > (1 << lg) ? lg + 1 : lg;
}

int logexact(int num) {
  //
  // return log to base 2 if number is an exact power of 2, log(else)=-1
  //
  int
    lg = log_down(num);
  return num > (1 << lg) ? -1 : lg;
}
};

#endif
