//
// The ZlibEngine is a Tiny Software (tm) project.
// You may use the code in this project without restriction.
// Contact markn@tiny.com for more information.
//

#if !defined( _ZLIBENGN_H )
#define _ZLIBENGN_H

extern "C" {
#include "zlib.h"
}

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#define input_length 65536
#define output_length 65536

class ZlibEngine : public z_stream {
    public :
        ZlibEngine();
        int compress(const char *input, const char *output, int level, int strategy);
        int decompress(const char *input, const char *output);
	char_vector huffman_encoder(char_vector&);
	char_vector encoder(char_vector&, int level);
	char_vector decoder(char_vector&);

//
// These three functions are only used internally.
//
    protected :
        int load_input();
        int flush_output();
        int load_buffer(char_vector&);
        int flush_buffer(char_vector&);
//
// The remaining data members are used internally
//
    protected :
        FILE *fin;
        FILE *fout;
        long length;
        int err;
	char_vector_iterator in_cv;
        unsigned char input_buffer[input_length];
        unsigned char output_buffer[output_length];
};

//
// I define one error code in addition to those
// found in zlib.h
//

#define Z_USER_ABORT (-7)

//
// The constructor initializes a couple of members
// of the z_stream class.  See the Zlib documentation
// for details on what those members do
//

ZlibEngine::ZlibEngine() {
    zalloc = 0;  // z_stream member
    zfree = 0;   // z_stream member
    opaque = 0;  // z_stream member
//
// I initialize these members just for tidiness.
//
    fin = 0;
    fout = 0;
}

int ZlibEngine::compress(const char *input, const char *output, int level, int strategy) {
    //
    // compress() is the public function used to compress
    // a single file.  It has to take care of opening the
    // input and output files and setting up the buffers for
    // Zlib.  It then calls deflate() repeatedly until all
    // input and output processing has been done, and finally
    // closes the files and cleans up the Zlib structures. 
    //
    struct stat
      st;
    err = Z_OK;
    avail_in = 0;
    avail_out = output_length;
    next_out = output_buffer;
    fin  = fopen(input, "rb");
    fout = fopen(output, "wb");
    length = st.st_size;
    deflateInit2(this, level, Z_DEFLATED, 15, 8, strategy);
    for (;;) {
        if (!load_input())
            break;
        err = deflate(this, Z_NO_FLUSH);
        flush_output();
        if (err != Z_OK)
            break;
    }
    for (;;) {
        err = deflate(this, Z_FINISH);
        if (!flush_output())
            break;
        if (err != Z_OK)
            break;
    }
    deflateEnd(this);
    if (err != Z_OK && err != Z_STREAM_END) {
         printf("Zlib Error\n");
	 exit(-1);
    } else {
        //printf("Success\n");
        err = Z_OK;
    }
    fclose(fin);
    fclose(fout);
    fin = 0;
    fout = 0;
    return err;
}


int ZlibEngine::decompress(const char *input, const char *output) {
    //
    // decompress has to do most of the same chores as compress().
    // The only major difference it has is the absence of the level
    // parameter.  The level isnt needed when decompressing data
    // using the deflate algorithm.
    //
    struct stat st;
    err = Z_OK;
    avail_in = 0;
    avail_out = output_length;
    next_out = output_buffer;
    fin  = fopen(input, "rb");
    fout = fopen(output, "wb");
    if (lstat(input, &st) < 0)
      exit(-1);
    length = st.st_size;
    inflateInit(this);
    for (;;) {
        if (!load_input())
            break;
        err = inflate(this, Z_NO_FLUSH);
        flush_output();
        if (err != Z_OK)
            break;
    }
    for (;;) {
        err = inflate(this, Z_FINISH);
        if (!flush_output())
            break;
        if (err != Z_OK)
            break;
    }
    inflateEnd(this);
    if (err != Z_OK && err != Z_STREAM_END) {
        printf("Zlib Error\n");
	exit(-1);
    } else {
        //printf("Success\n");
        err = Z_OK;
    }
    if (fin)
        fclose(fin);
    fin = 0;
    if (fout)
        fclose(fout);
    fout = 0;
    return err;
}

//
//  Every time Zlib consumes all of the data in the
//  input buffer, this function gets called to reload.
//  The avail_in member is part of z_stream, and is
//  used to keep track of how much input is available.
//
int ZlibEngine::load_input() {
    if (avail_in == 0) {
        next_in = input_buffer;
        avail_in = fread(input_buffer, 1, input_length, fin);
    }
    return avail_in;
}

int ZlibEngine::load_buffer(char_vector& src) {
    if (avail_in == 0) {
        next_in = input_buffer;
	avail_in = min(input_length, (int)(src.end() - in_cv));
	copy(in_cv, in_cv + avail_in, input_buffer);
	in_cv += avail_in;
    }
    return avail_in;
}

//
//  Every time Zlib filsl the output buffer with data,
//  this function gets called.  Its job is to write
//  that data out to the output file, then update
//  the z_stream member avail_out to indicate that more
//  space is now available.
//
int ZlibEngine::flush_output() {
    unsigned int count = output_length - avail_out;
    if (count) {
        if (fwrite(output_buffer, 1, count, fout) != count) {
            err = Z_ERRNO;
            return 0;
        }
        next_out = output_buffer;
        avail_out = output_length;
    }
    return count;
}

int ZlibEngine::flush_buffer(char_vector& dst) {
    unsigned int count = output_length - avail_out;
    if (count) {
        for (unsigned int i = 0; i < count; i++)
	  dst.push_back(output_buffer[i]);
        next_out = output_buffer;
        avail_out = output_length;
    }
    return count;
}

char_vector ZlibEngine::huffman_encoder(char_vector& text) {
    //
    // Encodes the vector text using Huffman ONLY
    //
    char_vector
      out;
    err = Z_OK;
    in_cv = text.begin();
    avail_in = 0;
    avail_out = output_length;
    next_out = output_buffer;
    length = text.size();
    deflateInit2(this, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15, 8, Z_HUFFMAN_ONLY);

    for (;;) {
        if (!load_buffer(text))
            break;
        err = deflate(this, Z_NO_FLUSH);
        flush_buffer(out);
        if (err != Z_OK)
            break;
    }
    for (;;) {
        err = deflate(this, Z_FINISH);
        if (!flush_buffer(out))
            break;
        if (err != Z_OK)
            break;
    }
    deflateEnd(this);
    if (err != Z_OK && err != Z_STREAM_END) {
        printf("Zlib Error\n");
	exit(-1);
    } else {
        //printf("Success\n");
        err = Z_OK;
    }
    //printf ("err=%d\n",err);
    return out;
}



char_vector ZlibEngine::decoder(char_vector& code) {
    //
    // Decodes the vector code
    //
    char_vector
      out;
    err = Z_OK;
    in_cv = code.begin();
    avail_in = 0;
    avail_out = output_length;
    next_out = output_buffer;
    length = code.size();
    inflateInit(this);
    for (;;) {
        if (!load_buffer(code))
            break;
        err = inflate(this, Z_NO_FLUSH );
        flush_buffer(out);
        if (err != Z_OK)
            break;
    }
    for (;;) {
        err = inflate(this, Z_FINISH);
        if (!flush_buffer(out))
            break;
        if (err != Z_OK)
            break;
    }
    inflateEnd(this);
    if (err != Z_OK && err != Z_STREAM_END) {
        printf("Zlib Error\n");
	exit(-1);
    } else {
        //printf("Success\n");
        err = Z_OK;
    }
    return out;
}



char_vector ZlibEngine::encoder(char_vector& text, int level) {
    //
    // Encodes the vector text using Huffman ONLY
    //
    char_vector
      out;
    err = Z_OK;
    in_cv = text.begin();
    avail_in = 0;
    avail_out = output_length;
    next_out = output_buffer;
    length = text.size();
    deflateInit(this, level);

    for (;;) {
        if (!load_buffer(text))
            break;
        err = deflate(this, Z_NO_FLUSH);
        flush_buffer(out);
        if (err != Z_OK)
            break;
    }
    for (;;) {
        err = deflate(this, Z_FINISH);
        if (!flush_buffer(out))
            break;
        if (err != Z_OK)
            break;
    }
    deflateEnd(this);
    if (err != Z_OK && err != Z_STREAM_END) {
        printf("Zlib Error\n");
	exit(-1);
    } else {
        //printf("Success\n");
        err = Z_OK;
    }
    //printf ("err=%d\n",err);
    return out;
}



char_vector rle_encoder(char_vector& text) {
  //
  // Run Length Encoding
  //
  if (text.size() == 0)
    return text;

  char_vector
    code;
  char_vector_iterator
    v_i = text.begin(),
    u_i;
  int
    len;

  while (v_i != text.end()) {
    u_i = v_i + 1;
    while ((u_i != text.end())&&(*u_i == *v_i))
      ++u_i;
    len = u_i - v_i;
    while (len > MAXCHAR) {
      code.push_back(*v_i);
      code.push_back((unsigned char)(MAXCHAR-1));
      len -= MAXCHAR;
    }
    code.push_back(*v_i);
    code.push_back((unsigned char)(len - 1));
    v_i = u_i;
  }
  return code;
}



char_vector rle_decoder(char_vector& code) {
  //
  // Run Length Decoding
  //
  char_vector
    text;
  unsigned int
    i;
  for (char_vector_iterator s_i = code.begin(); s_i != code.end(); s_i += 2)
    for (i = 0; i < 1 + (unsigned int)((*(s_i+1)) & 0xFF); i++)
      text.push_back(*s_i);
  return text;
}


char_vector mtf_encoder(char_vector& text) {
  //
  // move to front Encoding
  //
  if (text.size() == 0)
    return text;

  char_vector
    code;
  unsigned char
    order[MAXCHAR];

  int
    i, j;

  for (i = 0 ; i < MAXCHAR; i++)
    order[i] = (unsigned char)(i & 0xFF);
  for (char_vector_iterator c = text.begin(); c != text.end(); ++c) {
    for (i = 0 ; i < MAXCHAR; i++)
      if (order[i] == (*c & 0xff))
	break;
    code.push_back((unsigned char)(i & 0xFF));
    for (j = i ; j > 0 ; j--)
      order[j] = order[j - 1];
    order[0] = (unsigned char)(*c);
  }
  return code;
}



char_vector mtf_decoder(char_vector& code) {
  //
  // move to front decoding
  //
  if (code.size() == 0)
    return code;

  char_vector
    text;
  unsigned char
    c,
    order[MAXCHAR];
  int
    j;

  for (int i = 0 ; i < MAXCHAR; i++ )
    order[i] = (unsigned char)i;

  for (char_vector_iterator i = code.begin(); i != code.end(); ++i) {
    text.push_back((unsigned char)(order[*i] & 0xFF));
    c = order[*i];
    for (j = *i; j > 0 ; j--)
      order[j] = order[j - 1];
    order[0] = (unsigned char)c;
  }     
  return text;
}

#endif  // #if !defined(_ZLIBENGN_H)

